/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

typedef struct DATA_COUNT {
  FLOAT_T val;
  int count;
} *DATA_COUNT_PTR;

static int compare_float(const void *ap, const void *bp)
{
  FLOAT_T a = *(FLOAT_T *)ap;
  FLOAT_T b = *(FLOAT_T *)bp;

  if (a < b) return -1;
  if (a > b) return 1;
  return 0;
}

static int compare_data_count(const void *ap, const void *bp)
{
  struct DATA_COUNT a = *(struct DATA_COUNT *)ap;
  struct DATA_COUNT b = *(struct DATA_COUNT *)bp;

  if (a.val < b.val) return -1;
  if (a.val > b.val) return 1;
  return 0;
}

/* Binary search using index boundaries
 * -- the idea comes from Yusuke Shinyama's blog (June, 2009)
 *
 * Generally speaking, we first introduce boundaries between indices
 * of an array as follows:
 *
 *   boundary 0   1   2   3   4   5
 *   index      0   1   2   3   4
 *   key       "1" "3" "4" "8" "10"
 *
 * Then, for example, the binary search returns the boundaries lower=1,
 * upper=2 if we give "3" as a key (we have an exact match at index=1),
 * and returns lower=3, upper=3 if we give a key "7" (we have no exact
 * match, but these returned boundaries indicate that "7" falls between
 * indices 2 and 3).
 *
 * The function, bsearch_data_count(), implements this binary search
 * algorithm, being specialized to an array of data-count pairs.
 */
static void bsearch_data_count(unsigned int *lower, unsigned int *upper,
                               int pos, int size,
                               struct DATA_COUNT *data_count)
{
  int i,i0,i1,c;

  /* Do binary search */
  i0 = 0;
  i1 = size;
  while (i0 < i1) {
    i = (i0 + i1) / 2;
    c = data_count[i].count;
    if (pos < c)
      i1 = i;
    else if (pos > c)
      i0 = (i+1);
    else {
      i0 = i;
      i1 = (i+1);
      break;
    }
  }

  *lower = i0;
  *upper = i1;
}

FLOAT_T get_quantile(FLOAT_T q, int size, FLOAT_T *data)
{
  FLOAT_T pos;
  int pos_floor,pos_ceil;

  qsort(data, size, sizeof(FLOAT_T), compare_float);

  pos = q * (size - 1);  /* (1 - q + q * size) - 1 */
  if (pos < 0)
    die("Too small data (size = %d) for computing %.Nf-quantile\n", size, q);

  pos_floor = (int)pos;

  if (FABS(pos - pos_floor) < SMALL_FLOAT)
    return data[pos_floor];
  else {
    pos_ceil = pos_floor + 1;
    return
      (pos_ceil - pos) * data[pos_floor] + (pos - pos_floor) * data[pos_ceil];
  }
}

/* Elements in count[] are assumed to be positive */
FLOAT_T get_quantile_count(FLOAT_T q, int size, FLOAT_T *data, int *count)
{
  int i,i0,i1,total_count;
  int pos_floor,pos_ceil;
  unsigned int upper_ceil,lower_ceil,upper_floor,lower_floor;
  FLOAT_T pos,quantile;
  struct DATA_COUNT *data_count;

  data_count = (struct DATA_COUNT *)xmalloc(sizeof(struct DATA_COUNT)*size);
  for (i=0; i<size; i++) {
    data_count[i].val = data[i];
    data_count[i].count = count[i];
  }

  qsort(data_count, size, sizeof(struct DATA_COUNT), compare_data_count);

  for (i=1; i<size; i++)
    data_count[i].count += data_count[i - 1].count; /* cumulative */
  total_count = data_count[size - 1].count;

  pos = 1 - q + q * total_count;
  if (pos < 1)
    die("Too small data (size = %d) for computing %.Nf-quantile\n", size, q);

  pos_floor = (int)pos;
  pos_ceil =
    (FABS(pos - pos_floor) < SMALL_FLOAT) ? pos_floor : (pos_floor + 1);

  if (pos_floor == pos_ceil) {
    bsearch_data_count(&lower_floor,&upper_floor,pos_floor,size,data_count);
    i = lower_floor;
    quantile = data_count[i].val;
  }
  else {
    bsearch_data_count(&lower_floor,&upper_floor,pos_floor,size,data_count);
    bsearch_data_count(&lower_ceil, &upper_ceil, pos_ceil, size,data_count);

    if (lower_floor < upper_floor && lower_ceil < upper_ceil) {
      /* exact matches for both */
      i0 = lower_floor;
      i1 = lower_ceil;
      quantile =
        (pos_ceil - pos)  * data_count[i0].val +
        (pos - pos_floor) * data_count[i1].val;
    }
    else if (lower_floor < upper_floor && lower_ceil >= upper_ceil) {
      /* exact match only for the floor value */
      i0 = lower_floor;
      i1 = upper_ceil;
      quantile =
        (pos_ceil - pos)  * data_count[i0].val +
        (pos - pos_floor) * data_count[i1].val;
    }
    else if (lower_floor >= upper_floor && lower_ceil < upper_ceil) {
      /* exact match only for the ceil value */
      i0 = upper_floor;
      i1 = lower_ceil;
      quantile =
        (pos_ceil - pos)  * data_count[i0].val +
        (pos - pos_floor) * data_count[i1].val;
    }
    else if (lower_floor >= upper_floor && lower_ceil >= upper_ceil) {
      /* no exact match */
      i = upper_floor;
      quantile = data_count[i].val;
    }
  }

  xfree(data_count);

  return quantile;
}

#if 0
int main(int argc, char *argv[])
{
  FLOAT_T data[6]  = { 0.8, 1.5, 1.3, 1.1, 0.9, 1.9 };
  int count[6]     = {   2,   2,   4,   1,   2,   1 };
  int size = 6;
  FLOAT_T q;

  for (q=0.0; q<=1.0 + SMALL_FLOAT; q += 0.025) {
    printf("Q=%.3Lf | Quantile=%.3Lf\n",
           q, get_quantile_count(q,size,data,count));
  }
}
#endif
