/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

/*
 * [CSV form]
 *   - does not understand the double quotation character
 *   - allows the space/tab characters
 *   - ignores the lines starting with '#'
 *
 * [Tab-separated from]
 *   - does not understand the double quotation character
 *   - allows the space characters
 *   - regards the consecutive tab characteres as one tab character
 *   - ignores the lines starting with '#'
 *
 * [Syntax of the hyperparameter specification]
 *
 *   ## comment1 (could be multiple lines)
 *   c0,P(c0)
 *   c1,P(c1)
 *    :
 *   ## comment2-0 (discrete attr. -- attr ID, class, value, hparam)
 *   a0,c0,v0_0,P(v0_0|c0)
 *    :
 *   a0,c0,v0_N,P(v0_N|c0)
 *   a0,c1,v0_0,P(v0_0|c1)
 *    :
 *   ## comment2-1 (discrete attr.)
 *   a1,c0,v1_0,P(v1_0|c0)
 *    :
 *   ## comment2-2 (discrete attr.)
 *   a2,c0,v2_0,P(v2_0|c0)
 *    :
 *   ## comment3-0 (cont. attr. -- dist ID, class, attr ID, mean)
 *   d0,c0,a0_0,mean(0,0,0)
 *    :
 *   d0,c0,a0_M,mean(0,0,M)
 *   d0,c1,a0_0,mean(1,0,0)
 *    :
 *   ## comment3-1 (cont. attr.)
 *   d1,c0,a0_0,mean(0,1,0)
 *    :
 *   ## comment3-2 (cont. attr.)
 *   d2,c0,a0_0,mean(0,2,0)
 *    :
 *   ## comment4-0 (cont. attr. -- dist ID, class, attr ID, attr ID, covar)
 *   d0,c0,a0_0,a1_0,covar(0,0,0,0)
 *   d0,c0,a0_0,a1_1,covar(0,0,0,1)
 *    :
 *   d0,c0,a0_M,a1_M,covar(0,0,M,M)
 *   d0,c1,a0_0,a1_0,covar(1,0,0,0)
 *    :
 *   ## comment4-1 (cont. attr.)
 *   d1,c0,a0_0,a1_0,covar(0,1,0,0)
 *    :
 *   ## comment4-2 (cont. attr.)
 *   d2,c0,a0_0,a1_0,covar(0,2,0,0)
 *    :
 *   ## comment5 (cont. attr. -- class, tau)
 *   c0,tau(0)
 *   c1,tau(1)
 *    :
 *   ## comment6-0 (cont. attr. -- dist ID, class, alpha)
 *   d0,c0,alpha(0,0)
 *   d0,c1,alpha(1,0)
 *    :
 *   ## comment6-1 (cont. attr. -- dist ID, class, alpha)
 *   d1,c0,alpha(0,1)
 *   d1,c1,alpha(1,1)
 *    :
 *   ## comment6-2 (cont. attr. -- dist ID, class, alpha)
 *   d2,c0,alpha(0,2)
 *   d2,c1,alpha(1,2)
 *    :
 *
 * [Note]
 *   Basically, parameters/hyperparameters files are well-formed,
 *   so we do not output detailed error messages.
 */
static int read_hparam_body(char *fname, FILE *fp)
{
  char *buf = NULL, *buf_raw = NULL;
  char **word = NULL, **word_raw = NULL;
#ifdef USE_ICONV
  char *buf_out, *word_out;
  size_t buf_len_in, buf_len_out;
  size_t *w_len_in = NULL, *w_len_out = NULL;
  ICONV_BUF_TYPE buf_in, *word_in = NULL;
#endif
  int line_c,v;
  FLOAT_T param;
  int num_comment_field,is_comment,ret;
  char *endptr;
  int tmp_k=-1,prev_tmp_k,tmp_i,tmp_u;
  int tmp_v,tmp_v0,tmp_v1,tmp_a,tmp_a0,tmp_a1;
  FLOAT_T tmp_tau,tmp_alpha;
  int max_num_class,new_max_num_class;
  int *max_num_word = NULL;
  int *new_max_num_word = NULL;
  char **field_class = NULL, **field_cat = NULL;
  char **field_mean  = NULL, **field_covar = NULL;
  char **field_tau   = NULL, **field_alpha = NULL;

  buf_raw = alloc_str(MAX_LINE_LEN + 1);

  if (mode_nominal) {
    word_raw = (char **)xmalloc(sizeof(char *)*tuple_size_d);
    for (v=0; v<tuple_size_d; v++)
      word_raw[v] = alloc_str(MAX_WORD_LEN + 1);
  }

  if (mode_conv_in)
    buf = alloc_str(MAX_LINE_LEN + 1);

  if (mode_nominal) {
    if (mode_conv_out) {
      word = (char **)xmalloc(sizeof(char *)*tuple_size_d);
      for (v=0; v<tuple_size_d; v++)
        word[v] = alloc_str(MAX_WORD_LEN + 1);

#ifdef USE_ICONV
      word_in = (ICONV_BUF_TYPE *)xmalloc(sizeof(ICONV_BUF_TYPE)*tuple_size_d);

      for (v=0; v<tuple_size_d; v++)
        word_in[v] = alloc_str(MAX_WORD_LEN + 1);

      w_len_in = (size_t *)xmalloc(sizeof(size_t)*tuple_size_d);
      w_len_out = (size_t *)xmalloc(sizeof(size_t)*tuple_size_d);
#endif
    }
  }

  /* allocate the temporary areas */
  max_num_class = new_max_num_class = INIT_TMP_CLASS;
  max_num_word = new_max_num_word = NULL;

  if (mode_nominal) {
    max_num_word = (int *)xmalloc(sizeof(int)*tuple_size_d);
    new_max_num_word = (int *)xmalloc(sizeof(int)*tuple_size_d);
    for (v=0; v<tuple_size_d; v++)
      max_num_word[v] = new_max_num_word[v] = INIT_TMP_WORD;
  }

  alloc_read_hparam(max_num_class, max_num_word);

  line_c = 0;
  num_comment_field = 0;
  is_comment = 0;
  num_class = 0;

  if (mode_nominal) {
    num_word = (int *)xmalloc(sizeof(int)*tuple_size_d);
    for (v=0; v<tuple_size_d; v++)
      num_word[v] = 0;
  }

  field_class = alloc_field(NUM_FIELD_CLASS);
  if (mode_nominal)
    field_cat = alloc_field(NUM_FIELD_CAT);
  if (mode_gauss) {
    field_mean  = alloc_field(NUM_FIELD_MEAN);
    field_covar = alloc_field(NUM_FIELD_COVAR);
    field_tau   = alloc_field(NUM_FIELD_TAU);
    field_alpha = alloc_field(NUM_FIELD_ALPHA);
  }

  while (fgets(buf_raw,MAX_LINE_LEN,fp) != NULL) {
    line_c++;

#ifdef USE_ICONV
    if (mode_conv_in) {
      buf_in = buf_raw;
      buf_out = buf;
      buf_len_in  = strlen(buf_raw)+1;
      buf_len_out = MAX_LINE_LEN+1;
      if (iconv(cd_in,&buf_in,&buf_len_in,&buf_out,&buf_len_out) == (size_t)-1)
        die("[%s, line %d]"
            " Couldn't convert to the internal charcode (errno=%d)\n",
            fname,line_c,errno);
    }
    else buf = buf_raw;
#else
    buf = buf_raw;
#endif

    /* we count up the number of comment areas */
    if (buf[0] == '#') {
      is_comment = 1;
      continue;
    }
    else {
      if (is_comment) num_comment_field++;
      is_comment = 0;
    }

    /**** Parse data matrices, which are separated by the comment lines ****/

    if (num_comment_field >
        tuple_size_d + 3 * gauss_size + ((mode_gauss) ? 1 : 0) + 1) {
      die("Too many attribute hyperparameter fields in %s\n",fname_hparam);
    }
    /** Reading P(c) ... **/
    else if (num_comment_field == 1) {

      ret = -1;
      if (format_in == FMT_CSV)
        ret = split_comma(NUM_FIELD_CLASS,buf,field_class,NULL);
      else
        ret = split_tab(NUM_FIELD_CLASS,buf,field_class,NULL);
      if (ret != NUM_FIELD_CLASS)
        DIE_LINE_NFIELD(fname,line_c,buf,
                        "Unexpected # of fields for class probs");

      tmp_k = strtol(field_class[0],&endptr,10);
      if (*endptr != '\0') DIE_LINE_ILL_INT(fname,line_c,buf);

      if (num_class != tmp_k)
        die_line(fname,line_c,__LINE__,buf,
                 "Unexpected order of class numbers",-1);
      else
        num_class++;

      param = STRTOD(field_class[1],&endptr);
      if (*endptr != '\0') DIE_LINE_ILL_FLOAT(fname,line_c,buf);

      /* expand the temporary areas if necessary */
      if (tmp_k >= max_num_class) {
        new_max_num_class = max_num_class * 2;
        expand_read_hparam_c(max_num_class, new_max_num_class);
        if (mode_nominal)
          expand_read_hparam_w(max_num_class, new_max_num_class,
                               max_num_word, max_num_word);
        max_num_class = new_max_num_class;
      }

      read_hparam_c[tmp_k] = param;
    }
    /** Reading P(x|k) ... **/
    else if (num_comment_field > 1 &&
             num_comment_field <= tuple_size_d + 1) {

      v = num_comment_field - 2; /* attribute ID */

      if (format_in == FMT_CSV)
        ret = split_comma(NUM_FIELD_CAT,buf,field_cat,NULL);
      else
        ret = split_tab(NUM_FIELD_CAT,buf,field_cat,NULL);
      if (ret != NUM_FIELD_CAT)
        DIE_LINE_NFIELD(fname,line_c,buf,
                        "Unexpected # of fields for discrete attribute probs");

      tmp_v = strtol(field_cat[0],&endptr,10);
      if (*endptr != '\0')
        DIE_LINE_ILL_INT(fname,line_c,buf);

      if (tmp_v < 0 || tmp_v >= tuple_size || v != attr_info[tmp_v].local_id)
        die_line(fname,line_c,__LINE__,buf,"Unexpected attribute ID",-1);

      prev_tmp_k = tmp_k;
      tmp_k = strtol(field_cat[1],&endptr,10);
      if (*endptr != '\0') DIE_LINE_ILL_INT(fname,line_c,buf);

      /* from the syntax of the hyperparameter file, the number of class
       * has been already determined here.
       */
      if (tmp_k >= num_class)
        die_line(fname,line_c,__LINE__,buf,
                 "Different number of classes exist in entries for P(c)"
                 " and P(x1|c)\n",
                 -1);

      if (tmp_k != prev_tmp_k)
        num_word[v] = 0;  /* reset the number of attribute values */

      if (strlen(field_cat[2]) == 0)
        die_line(fname,line_c,__LINE__,buf,"Empty value",-1);

      strncpy(word_raw[v],field_cat[2],MAX_WORD_LEN);

#ifdef USE_ICONV
      if (mode_conv_out) {
        word_in[v] = word_raw[v];
        word_out = word[v];
        w_len_in[v] = strlen(word_raw[v])+1;
        w_len_out[v] = MAX_WORD_LEN+1;
        if (iconv(cd_out,&word_in[v],&w_len_in[v],&word_out,&w_len_out[v]) == (size_t)-1)
          die("[%s, line %d] Couldn't convert to %s (errno=%d)\n",
              fname,line_c,charcode_out,errno);
      }
      else word = word_raw;
#else
      word = word_raw;
#endif

      tmp_i = register_hashtable(&ht[v],word[v]);
      if (num_word[v] != tmp_i)
        die("[%s, line %d] Found an inconsistent entry (%s)",
            fname,line_c,word[v]);
      else
        num_word[v]++;

      param = STRTOD(field_cat[3],&endptr);
      if (*endptr != '\0') DIE_LINE_ILL_FLOAT(fname,line_c,buf);

      if (tmp_k >= max_num_class || tmp_i >= max_num_word[v]) {
        new_max_num_class =
          (tmp_k >= max_num_class) ? max_num_class * 2 : max_num_class;
        new_max_num_word[v] =
          (tmp_i >= max_num_word[v]) ? max_num_word[v] * 2 : max_num_word[v];
        expand_read_hparam_w(max_num_class, new_max_num_class,
                             max_num_word, new_max_num_word);
        max_num_class = new_max_num_class;
        memcpy(max_num_word,new_max_num_word,sizeof(int)*tuple_size_d);
      }

      read_hparam_w[tmp_k][v][tmp_i] = param;
    }
    /** Reading the means of P(x|k) ... **/
    else if (num_comment_field > tuple_size_d + 1 &&
             num_comment_field <= tuple_size_d + 2*gauss_size + 1 &&
             (num_comment_field - tuple_size_d - 2) % 2 == 0) {

      if (!mode_gauss)
        die_line(fname, line_c, __LINE__, buf,
                 "Unexpected hyperparameters for Gaussians", -1);

      if (format_in == FMT_CSV)
        ret = split_comma(NUM_FIELD_MEAN,buf,field_mean,NULL);
      else
        ret = split_tab(NUM_FIELD_MEAN,buf,field_mean,NULL);
      if (ret != NUM_FIELD_MEAN)
        DIE_LINE_NFIELD(fname,line_c,buf,
                        "Unexpected # of fields for means of"
                        " continuous attributes");

      tmp_u = strtol(field_mean[0],&endptr,10);
      if (*endptr != '\0') DIE_LINE_ILL_INT(fname,line_c,buf);

      tmp_k = strtol(field_mean[1],&endptr,10);
      if (*endptr != '\0') DIE_LINE_ILL_INT(fname,line_c,buf);

      tmp_v = strtol(field_mean[2],&endptr,10);
      if (*endptr != '\0') DIE_LINE_ILL_INT(fname,line_c,buf);

      param = STRTOD(field_mean[3],&endptr);
      if (*endptr != '\0') DIE_LINE_ILL_FLOAT(fname,line_c,buf);

      tmp_a = attr_info[tmp_v].local_id;

      read_hparam_u[tmp_k][tmp_u][tmp_a] = param;
    }
    /** Reading the covariances of P(x|k) ... **/
    else if (num_comment_field > tuple_size_d + 1 &&
             num_comment_field <= tuple_size_d + 2*gauss_size + 1 &&
             (num_comment_field - tuple_size_d - 2) % 2 == 1 ) {

      if (!mode_gauss)
        die_line(fname, line_c, __LINE__, buf,
                 "Unexpected hyperparameters for Gaussians", -1);

      if (format_in == FMT_CSV)
        ret = split_comma(NUM_FIELD_COVAR,buf,field_covar,NULL);
      else
        ret = split_tab(NUM_FIELD_COVAR,buf,field_covar,NULL);
      if (ret != NUM_FIELD_COVAR)
        DIE_LINE_NFIELD(fname,line_c,buf,
                        "Unexpected # of fields for covariances of"
                        " continuous attributes");

      tmp_u = strtol(field_covar[0],&endptr,10);
      if (*endptr != '\0') DIE_LINE_ILL_INT(fname,line_c,buf);

      tmp_k = strtol(field_covar[1],&endptr,10);
      if (*endptr != '\0') DIE_LINE_ILL_INT(fname,line_c,buf);

      tmp_v0 = strtol(field_covar[2],&endptr,10);
      if (*endptr != '\0') DIE_LINE_ILL_INT(fname,line_c,buf);

      tmp_v1 = strtol(field_covar[3],&endptr,10);
      if (*endptr != '\0') DIE_LINE_ILL_INT(fname,line_c,buf);

      param = STRTOD(field_covar[4],&endptr);
      if (*endptr != '\0') DIE_LINE_ILL_FLOAT(fname,line_c,buf);

      tmp_a0 = attr_info[tmp_v0].local_id;
      tmp_a1 = attr_info[tmp_v1].local_id;

      read_hparam_s[tmp_k][tmp_u][tmp_a0][tmp_a1] = param;
    }
    /** Reading Tau's in the Normal-Wishart distribution... **/
    else if (num_comment_field == tuple_size_d + 2*gauss_size + 2) {

      if (!mode_gauss)
        die_line(fname, line_c, __LINE__, buf,
                 "Unexpected hyperparameters for Gaussians",-1);

      if (format_in == FMT_CSV)
        ret = split_comma(NUM_FIELD_TAU,buf,field_tau,NULL);
      else
        ret = split_tab(NUM_FIELD_TAU,buf,field_tau,NULL);
      if (ret != NUM_FIELD_TAU)
        DIE_LINE_NFIELD(fname,line_c,buf,
                        "Unexpected # of fields for tau's in"
                        " Normal Wishart distrib.");

      tmp_k = strtol(field_tau[0],&endptr,10);
      if (*endptr != '\0') DIE_LINE_ILL_INT(fname,line_c,buf);

      tmp_tau = STRTOD(field_tau[1],&endptr);
      if (*endptr != '\0') DIE_LINE_ILL_FLOAT(fname,line_c,buf);

      read_tau[tmp_k] = tmp_tau;
    }
    /** Reading Alpha's in the Normal-Wishart distribution... **/
    else if (num_comment_field >= tuple_size_d + 2*gauss_size + 3) {

      if (!mode_gauss)
        die_line(fname, line_c, __LINE__, buf,
                 "Unexpected hyperparameters for Gaussians", -1);

      if (format_in == FMT_CSV)
        ret = split_comma(NUM_FIELD_ALPHA,buf,field_alpha,NULL);
      else
        ret = split_tab(NUM_FIELD_ALPHA,buf,field_alpha,NULL);
      if (ret != NUM_FIELD_ALPHA)
        DIE_LINE_NFIELD(fname,line_c,buf,
                        "Unexpected # of fields for alpha's"
                        " in Normal Wishart distrib.");

      tmp_u = strtol(field_alpha[0],&endptr,10);
      if (*endptr != '\0') DIE_LINE_ILL_INT(fname,line_c,buf);

      tmp_k = strtol(field_alpha[1],&endptr,10);
      if (*endptr != '\0') DIE_LINE_ILL_INT(fname,line_c,buf);

      tmp_alpha = STRTOD(field_alpha[2],&endptr);
      if (*endptr != '\0') DIE_LINE_ILL_FLOAT(fname,line_c,buf);

      read_alpha_w[tmp_k][tmp_u] = tmp_alpha;
    }
  }

  release_field(field_class,NUM_FIELD_CLASS);
  if (mode_nominal)
    release_field(field_cat,NUM_FIELD_CAT);
  if (mode_gauss) {
    release_field(field_mean, NUM_FIELD_MEAN);
    release_field(field_covar,NUM_FIELD_COVAR);
    release_field(field_tau,  NUM_FIELD_TAU);
    release_field(field_alpha,NUM_FIELD_ALPHA);
  }

  if (num_comment_field <
      tuple_size_d + 3 * gauss_size + ((mode_gauss) ? 1 : 0) + 1)
    die("Insufficiennt hyperparameter specifications in %s"
        " (#data-matrices = %d)\n",
        fname_hparam,num_comment_field);

  if (mode_nominal) {
    for (v=0; v<tuple_size_d; v++)
      num_word[v] = get_cardinality(&ht[v]);
  }

  return max_num_class;
}

static void copy_read_hparam(void)
{
  int k,v,i,u,a,a0,a1,n;

  for (k=0; k<num_class; k++) {
    if (read_hparam_c[k] < -1)
      die(conv_msg("Illegal class hyperparameter: %.Nf (class=%d)\n"),
          read_hparam_c[k],k);
    hparam_c[k] = read_hparam_c[k];
  }

  if (mode_nominal) {
    for (k=0; k<num_class; k++)
      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++) {
          if (read_hparam_w[k][v][i] < -1)
            die(conv_msg("Illegal attribute hyperparameter: "
                         "%.Nf (class=%d, attribute=%d)\n"),
                read_hparam_w[k][v][i],k,v);
          hparam_w[k][v][i] = read_hparam_w[k][v][i];
        }
  }

  if (mode_gauss) {
    for (k=0; k<num_class; k++) {
      hparam_tau[k] = read_tau[k];

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];

        for (a=0; a<n; a++)
          hparam_u[k][u][a] = read_hparam_u[k][u][a];

        for (a0=0; a0<n; a0++)
          for (a1=a0; a1<n; a1++)
            hparam_s[k][u][a0][a1] = read_hparam_s[k][u][a0][a1];
        copy_upper_to_lower(hparam_s[k][u]);

        hparam_alpha_w[k][u] = read_alpha_w[k][u];
      }
    }

    for (k=0; k<num_class; k++)
      for (u=0; u<gauss_size; u++)
        hparam_sdet[k][u] = INVERT(hparam_s[k][u],hparam_p[k][u]);
  }
}

/*
 * [Note] We also use read_param_body() and release_read_param() for
 * reading hyperparameters.
 */
void read_hparam(void)
{
  int v,i;
  int max_num_class;
  FILE *fp_hparam;

  print_msg("Reading hyperparameters from %s...", fname_hparam);
  fflush(fp_msg);

  if (mode_nominal) {
    ht = (HASHTABLE *)xmalloc(sizeof(HASHTABLE)*tuple_size_d);

    for (v=0; v<tuple_size_d; v++)
      init_hashtable(&ht[v]);

    num_word = (int *)xmalloc(sizeof(int)*tuple_size_d);
  }

  fp_hparam = fopen(fname_hparam,"rb");
  if (fp_hparam == NULL) die("Couldn't open %s\n", fname_hparam);

  /* alloc_read_hparam() is called */
  max_num_class = read_hparam_body(fname_hparam,fp_hparam);

  alloc_hparam();
  if (mode_gauss) {
    alloc_hparam_gauss();
    alloc_hparam_gauss_aux();
  }

  copy_read_hparam();
  release_read_hparam(max_num_class);

#ifndef VNBC
  if (mode_verbose) print_hparam(fp_log);
#endif

  if (mode_nominal) {
    word_list = (char ***)xmalloc(sizeof(char **)*tuple_size_d);

    for (v=0; v<tuple_size_d; v++)
      word_list[v] = get_keys(&ht[v]);

    print_order = (int **)xmalloc(sizeof(int *)*tuple_size_d);

    for (v=0; v<tuple_size_d; v++) {
      print_order[v] = (int *)xmalloc(sizeof(int)*num_word[v]);
      for (i=0; i<num_word[v]; i++) {
        print_order[v][i] = i;
      }
      if (mode_lex_order) {
        tmp_word_list = word_list[v];
        qsort(print_order[v],num_word[v],sizeof(int),
              (int(*)(const void *,const void *))compare_word);
      }
    }
  }

  if (mode_verbose) print_attr_list(fp_log);

  print_msg("done\n");
}
