/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

static void check_students_t(int n, FLOAT_T d)
{
  if (ISNAN(d)) {
    print_param(fp_log);
    DIE_NUMERIC("Student's-t p.d.f. returned NaN\n");
  }
  if (ISINF(d)) {
    print_param(fp_log);
    DIE_NUMERIC("Student's-t p.d.f. returned INF\n");
  }
}

static FLOAT_T lstudents_t_uv_const(FLOAT_T sigma, FLOAT_T f)
{
  FLOAT_T d = 0.0;

  d = - 0.5 * LOG(f * M_PI) + LGAMMA(0.5 * (f + 1)) - LGAMMA(0.5 * f)
      - 0.5 * LOG(sigma);

  return d;
}

static FLOAT_T lstudents_t_mv_const(FLOAT_T n, MATRIX sigma, FLOAT_T f)
{
  FLOAT_T d = 0.0;

  d -= 0.5 * n * LOG(f * M_PI);
  d += LGAMMA(0.5 * (f + n));
  d -= LGAMMA(0.5 * f);
  d -= 0.5 * LOG(DETERMINANT(sigma));

  return d;
}

static FLOAT_T lstudents_t_uv(FLOAT_T x, FLOAT_T mu,
                              FLOAT_T sigma, FLOAT_T f)
{
  FLOAT_T d;

  d = lstudents_t_uv_const(sigma,f);
  d -= 0.5 * (f + 1) * LOG(1 + (x - mu) * (x - mu) / (f * sigma));

  return d;
}

static FLOAT_T
lstudents_t_mv(int n, VECTOR x, VECTOR mu, MATRIX sigma, FLOAT_T f)
{
  FLOAT_T d;
  MATRIX sigma_inv;
  VECTOR x0;

  d = lstudents_t_mv_const(n,sigma,f);

  x0 = dup_vector(x);
  sigma_inv = alloc_matrix(n,n);

  diff_vector(x0,mu,NULL);
  INVERT(sigma,sigma_inv);
  div_const_matrix(sigma_inv,f);

  d -= 0.5 * (f + n) * LOG(1.0 + qform(sigma_inv,x0));

  release_vector(x0);
  release_matrix(sigma_inv);

  return d;
}

FLOAT_T lstudents_t(VECTOR x, VECTOR mu, MATRIX sigma, FLOAT_T f)
{
  int n;
  FLOAT_T d;

#ifdef CHECK_DIM
  if ((n = dim(x)) <= 0 || n != dim(mu) || n != dim0(sigma) || n != dim1(sigma))
    DIE_DIM;
#else
  n = dim(x);
#endif

  if (n == 1)
    d = lstudents_t_uv(x[0],mu[0],sigma[0][0],f);
  else
    d = lstudents_t_mv(n,x,mu,sigma,f);

  check_students_t(n,d);

  return d;
}

FLOAT_T students_t(VECTOR x, VECTOR mu, MATRIX sigma, FLOAT_T f)
{
  return EXP(lstudents_t(x,mu,sigma,f));
}

FLOAT_T students_t_uv(FLOAT_T x, FLOAT_T mu, FLOAT_T sigma, FLOAT_T f)
{
  return EXP(lstudents_t_uv(x,mu,sigma,f));
}

FLOAT_T
lstudents_t_marginal(VECTOR x, VECTOR mu, MATRIX sigma, FLOAT_T f, int *obs)
{
  int n,n_obs,n_miss,i;
  VECTOR x_o,mu_o;
  MATRIX s_oo,s_om,s_mo,s_mm,s_mm_inv;
  MATRIX z_oo,z_om,sigma_oo;
  int *miss;
  FLOAT_T d;

#ifdef CHECK_DIM
  if ((n = dim(x)) <= 0 || n != dim(mu) || n != dim0(sigma) || n != dim1(sigma))
    DIE_DIM;
#else
  n = dim(x);
#endif

  miss = (int *)xmalloc(sizeof(int)*n);
  n_obs = 0;
  for (i=0; i<n; i++) {
    if (obs[i]) {
      n_obs++;
      miss[i] = 0;
    }
    else miss[i] = 1;
  }
  n_miss = n - n_obs;

  if (n_miss == 0 || n_obs == 0)
    print_msg("n_miss = %d, n_obs = %d\n",n_miss,n_obs);

  x_o = alloc_proj_vector(x,obs);
  mu_o = alloc_proj_vector(mu,obs);

  s_oo = alloc_proj_matrix_n(sigma,n_obs,n_obs,obs,obs);
  s_om = alloc_proj_matrix_n(sigma,n_obs,n_miss,obs,miss);
  s_mo = alloc_proj_matrix_n(sigma,n_miss,n_obs,miss,obs);
  s_mm = alloc_proj_matrix_n(sigma,n_miss,n_miss,miss,miss);
  s_mm_inv = alloc_matrix(n_miss,n_miss);
  z_oo = alloc_matrix(n_obs,n_obs);
  z_om = alloc_matrix(n_obs,n_miss);
  sigma_oo = alloc_matrix(n_obs,n_obs);

  INVERT(s_mm,s_mm_inv);
  if (err_inv)
    DIE_NUMERIC("Failed to invert the covariances of the missing part\n");
  prod_matrix(s_om,s_mm_inv,z_om);
  prod_matrix(z_om,s_mo,z_oo);
  diff_matrix(s_oo,z_oo,sigma_oo);

  d = lstudents_t(x_o,mu_o,sigma_oo,f);

  release_vector(x_o);
  release_vector(mu_o);
  release_matrix(s_oo);
  release_matrix(s_om);
  release_matrix(s_mo);
  release_matrix(s_mm);
  release_matrix(s_mm_inv);
  release_matrix(z_oo);
  release_matrix(z_om);
  release_matrix(sigma_oo);

  return d;
}

FLOAT_T
students_t_marginal(VECTOR x, VECTOR mu, MATRIX sigma, FLOAT_T f, int *obs)
{
  return EXP(lstudents_t_marginal(x,mu,sigma,f,obs));
}

