/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "xgetopt.h"

void init_xgetopt(void)
{
  xoptind = 1;
}

/* Platform-independent version of snprintf() */
static int xsnprintf(char *s, size_t size, const char *fmt, ...)
{
  va_list ap;
  int ret;

  va_start(ap,fmt);
#ifdef _MSC_VER
  ret = _vsnprintf(s,(size - 1),fmt,ap);
  s[size - 1] = '\0';
#else
  ret = vsnprintf(s,size,fmt,ap);
#endif
  va_end(ap);

  return ret;
}

static int find_short_option(char c, OPTION *opts, int *req_arg)
{
  int i = 0;
  while (opts[i].id >= 0) {
    if (c == opts[i].short_opt) {
      *req_arg = opts[i].req_arg;
      return opts[i].id;
    }
    i++;
  }
  return -1;
}

static int find_long_option(char *s, OPTION *opts, int *req_arg)
{
  int i = 0;
  while (opts[i].id >= 0) {
    if (strcmp(s,opts[i].long_opt) == 0) {
      *req_arg = opts[i].req_arg;
      return opts[i].id;
    }
    i++;
  }
  return -1;
}

static void set_xoptarg(char *s)
{
  strncpy(xoptarg,s,OPTARG_LEN);
  strncpy(xoptarg_c,xoptarg,OPTARG_LEN);
}

static void clear_optstr(void)
{
  memset(optstr,0,OPTSTR_LEN);
}

static void clear_xoptarg(void)
{
  memset(xoptarg,0,OPTARG_LEN);
  memset(xoptarg_c,0,OPTARG_LEN);
}

int xgetopt(int argc, char *argv[], OPTION *opts)
{
  int len,req_arg,id;
  char *s, *arg_s;

  clear_optstr();
  clear_xoptarg();

  if (xoptind == argc) return GETOPT_FIN;

  s = argv[xoptind];
  strncpy(optstr,s,OPTSTR_LEN);
  len = strlen(s);
  xoptind++;

  if (len > 2 && s[0] == '-' && s[1] == '-' && isalpha(s[2])) { /* long opt */
    id = find_long_option((s + 2),opts,&req_arg);
    if (id < 0) return GETOPT_ERR;
  }
  else if (len > 2 && s[0] == '-' && isalpha(s[1])) { /* illegal long opt */
    return GETOPT_ERR;
  }
  else if (len == 2 && s[0] == '-' && isalpha(s[1])) { /* short opt */
    id = find_short_option(s[1],opts,&req_arg);
    if (id < 0) return GETOPT_ERR;
  }
  else { /* usual arguments */
    set_xoptarg(s);
    return 0;
  }

  if (req_arg) { /* option argument */
    if (xoptind == argc) return GETOPT_ERR;
    arg_s = argv[xoptind];
    if ((strlen(arg_s) > 1 && arg_s[0] == '-' && isalpha(arg_s[1]))
        || (strlen(arg_s) > 2 && arg_s[0] == '-' && arg_s[1] == '-'
            && isalpha(arg_s[2]))) {
      return GETOPT_ERR;
    }
    set_xoptarg(arg_s);
    xoptind++;
  }
  else {
    clear_xoptarg();
  }

  return id;
}

/* [Note] The returned string can (should) be freed by the caller */
char *get_option_str(int id, OPTION *opts, int long_only)
{
  int i,len;
  char *s;

  i = 0;
  while (opts[i].id >= 0) {
    if (opts[i].id == id) {
      if (opts[i].short_opt > 0 && !long_only) {
        len = strlen(opts[i].long_opt) + 8;
        s = (char *)malloc(sizeof(char)*len);
        if (s == NULL) {
          fprintf(stderr,"Couldn't allocate memory (get_option_str)\n");
          exit(1);
        }
        xsnprintf(s,len,"--%s (-%c)",opts[i].long_opt,opts[i].short_opt);
      }
      else {
        len = strlen(opts[i].long_opt) + 3;
        s = (char *)malloc(sizeof(char)*len);
        if (s == NULL) {
          fprintf(stderr,"Couldn't allocate memory (get_option_str)\n");
          exit(1);
        }
        xsnprintf(s,len,"--%s",opts[i].long_opt);
      }
      return s;
    }
    i++;
  }

  return NULL;
}

/* We use a safer version of strncpy() since the allocated area of
 * xoptarg might be larger than the allocated area for the destination.
 */
char *copy_optarg(char *dst, size_t n)
{
  strncpy(dst,xoptarg,n);
  dst[n] = '\0';

  return dst;
}
